@@ -1,4 +1,7 @@
-/*  Copyright (C) 1995, 1997, 1999, 2000-2002 Free Software Foundation, Inc.
+/* Replacement for GNU C library function getline
+
+   Copyright (C) 1995, 1997, 1999, 2000, 2001, 2002, 2003 Free
+   Software Foundation, Inc.
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -20,22 +23,16 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 # include <stddef.h>
 # include <stdio.h>
 
-# ifndef PARAMS
-#  if defined PROTOTYPES || (defined __STDC__ && __STDC__)
-#   define PARAMS(Args) Args
-#  else
-#   define PARAMS(Args) ()
-#  endif
-# endif
+/* Get ssize_t.  */
+# include <sys/types.h>
 
 /* glibc2 has these functions declared in <stdio.h>.  Avoid redeclarations.  */
 # if __GLIBC__ < 2
 
-int
-getline PARAMS ((char **_lineptr, size_t *_n, FILE *_stream));
+extern ssize_t getline (char **_lineptr, size_t *_linesize, FILE *_stream);
 
-int
-getdelim PARAMS ((char **_lineptr, size_t *_n, int _delimiter, FILE *_stream));
+extern ssize_t getdelim (char **_lineptr, size_t *_linesize, int _delimiter,
+			  FILE *_stream);
 
 # endif
 
