@@ -25,6 +25,15 @@
 extern int errno;
 #endif
 
+
+#if defined _MSC_VER || defined __MINGW32__
+/* Native Woe32 API.  */
+#include <process.h>
+#else
+/* Unix API.  */
+#include "wait.h"
+#endif
+
 #define WAITPID_CHILDREN 8
 static pid_t waited_pid[WAITPID_CHILDREN];
 static int waited_status[WAITPID_CHILDREN];
